/*
    This is a romfs file system driver for Windows NT and Windows 2000,
    pre-release 2000-07-11.
    Copyright (C) 2000 Bo Branten
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

NTSTATUS
FsdBuildRequest (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )
{
    BOOLEAN             IsTopLevelIrp = FALSE;
    PFSD_IRP_CONTEXT    IrpContext = NULL;
    NTSTATUS            Status = STATUS_UNSUCCESSFUL;

    try
    {
        try
        {
            FsdDbgPrintCall(DeviceObject, Irp);

            FsRtlEnterFileSystem();

            if (!IoGetTopLevelIrp())
            {
                IsTopLevelIrp = TRUE;
                IoSetTopLevelIrp(Irp);
            }

            IrpContext = FsdAllocateIrpContext(DeviceObject, Irp);

            if (!IrpContext)
            {
                Status = STATUS_INSUFFICIENT_RESOURCES;
                Irp->IoStatus.Status = Status;
                FsdCompleteRequest(Irp, IO_NO_INCREMENT);
            }
            else
            {
                Status = FsdDispatchRequest(IrpContext);
            }
        }
        except (FsdExceptionFilter(IrpContext, GetExceptionInformation()))
        {
            Status = FsdExceptionHandler(IrpContext, Irp);
        }
    }
    finally
    {
        if (IsTopLevelIrp)
        {
            IoSetTopLevelIrp(NULL);
        }

        FsRtlExitFileSystem();
    }
    
    return Status;
}

NTSTATUS
FsdQueueRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    // IsSynchronous means we can block (so we don't requeue it)
    IrpContext->IsSynchronous = TRUE;

    IoMarkIrpPending(IrpContext->Irp);

    ExInitializeWorkItem(
        &IrpContext->WorkQueueItem,
        FsdDeQueueRequest,
        IrpContext
        );

    ExQueueWorkItem(&IrpContext->WorkQueueItem, CriticalWorkQueue);

    return STATUS_PENDING;
}

VOID
FsdDeQueueRequest (
    IN PVOID Context
    )
{
    PFSD_IRP_CONTEXT IrpContext;

    IrpContext = (PFSD_IRP_CONTEXT) Context;

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    try
    {
        try
        {
            FsRtlEnterFileSystem();

            if (!IrpContext->IsTopLevel)
            {
                IoSetTopLevelIrp((PIRP) FSRTL_FSP_TOP_LEVEL_IRP);
            }

            FsdDispatchRequest(IrpContext);
        }
        except (FsdExceptionFilter(IrpContext, GetExceptionInformation()))
        {
            FsdExceptionHandler(IrpContext, IrpContext->Irp);
        }
    }
    finally
    {
        IoSetTopLevelIrp(NULL);

        FsRtlExitFileSystem();
    }
}

NTSTATUS
FsdDispatchRequest (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    switch (IrpContext->MajorFunction)
    {
    case IRP_MJ_CREATE:
        return FsdCreate(IrpContext);

    case IRP_MJ_CLOSE:
        return FsdClose(IrpContext);

    case IRP_MJ_READ:
        return FsdRead(IrpContext);

    case IRP_MJ_QUERY_INFORMATION:
        return FsdQueryInformation(IrpContext);

    case IRP_MJ_SET_INFORMATION:
        return FsdSetInformation(IrpContext);

    //case IRP_MJ_QUERY_EA:

    case IRP_MJ_FLUSH_BUFFERS:
        return FsdFlushBuffers(IrpContext);

    case IRP_MJ_QUERY_VOLUME_INFORMATION:
        return FsdQueryVolumeInformation(IrpContext);

    case IRP_MJ_DIRECTORY_CONTROL:
        return FsdDirectoryControl(IrpContext);

    case IRP_MJ_FILE_SYSTEM_CONTROL:
        return FsdFileSystemControl(IrpContext);

    case IRP_MJ_DEVICE_CONTROL:
        return FsdDeviceControl(IrpContext);

    case IRP_MJ_SHUTDOWN:
        return FsdShutdown(IrpContext);

    //case IRP_MJ_LOCK_CONTROL:

    case IRP_MJ_CLEANUP:
        return FsdCleanup(IrpContext);

    //case IRP_MJ_QUERY_SECURITY:
    //case IRP_MJ_QUERY_QUOTA:

#ifndef FSD_RO

    //case IRP_MJ_WRITE:
    //case IRP_MJ_SET_EA:
    //case IRP_MJ_SET_VOLUME_INFORMATION:
    //case IRP_MJ_SET_SECURITY:
    //case IRP_MJ_SET_QUOTA:

#endif // FSD_RO

    default:
        KdPrint((
            DRIVER_NAME ": FsdDispatchRequest: Unknown MajorFunction 0x%x\n",
            IrpContext->MajorFunction
            ));

        IrpContext->Irp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
        FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
        FsdFreeIrpContext(IrpContext);

        return STATUS_DRIVER_INTERNAL_ERROR;
    }
}

NTSTATUS
FsdExceptionFilter (
    IN PFSD_IRP_CONTEXT     IrpContext,
    IN PEXCEPTION_POINTERS  ExceptionPointers
    )
{
    NTSTATUS ExceptionCode;
    NTSTATUS Status;

    //
    // Only use a valid IrpContext
    //
    if (IrpContext)
    {
        if ((IrpContext->Identifier.Type != ICX) ||
            (IrpContext->Identifier.Size != sizeof(FSD_IRP_CONTEXT)))
        {
            IrpContext = NULL;
        }
    }

    ExceptionCode = ExceptionPointers->ExceptionRecord->ExceptionCode;

    if ((ExceptionCode == STATUS_IN_PAGE_ERROR) &&
        (ExceptionPointers->ExceptionRecord->NumberParameters >= 3))
    {
        ExceptionCode =
            ExceptionPointers->ExceptionRecord->ExceptionInformation[2];
    }

    //
    // If the exception is expected execute our handler
    //
    if (FsRtlIsNtstatusExpected(ExceptionCode))
    {
        KdPrint((DRIVER_NAME\
            ": FsdExceptionFilter: Catching exception 0x%x\n",
            ExceptionCode));

        Status = EXCEPTION_EXECUTE_HANDLER;

        if (IrpContext)
        {
            IrpContext->ExceptionInProgress = TRUE;
            IrpContext->ExceptionCode = ExceptionCode;
        }
    }
    //
    // else continue search for an higher level exception handler
    //
    else
    {
        KdPrint((DRIVER_NAME\
            ": FsdExceptionFilter: Passing on exception 0x%x\n",
            ExceptionCode));

        Status = EXCEPTION_CONTINUE_SEARCH;

        if (IrpContext)
        {
            FsdFreeIrpContext(IrpContext);
        }
    }

    return Status;
}

NTSTATUS
FsdExceptionHandler (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PIRP             Irp
    )
{
    NTSTATUS Status;

    //
    // Complete the IRP with the exception code as status
    //

    if (IrpContext)
    {
        Status = IrpContext->ExceptionCode;
        FsdFreeIrpContext(IrpContext);
    }
    else
    {
        Status = STATUS_INSUFFICIENT_RESOURCES;
    }

    Irp->IoStatus.Status = Status;
    FsdCompleteRequest(Irp, IO_NO_INCREMENT);

    return Status;
}

VOID
FsdCharToWchar (
    IN OUT PWCHAR   Destination,
    IN PCHAR        Source,
    IN ULONG        Length
    )
{
    ULONG Index;

    ASSERT(Destination != NULL);
    ASSERT(Source != NULL);

    for (Index = 0; Index < Length; Index++)
    {
        Destination[Index] = (WCHAR) Source[Index];
    }
}

#ifndef FSD_RO

NTSTATUS
FsdWcharToChar (
    IN OUT PCHAR    Destination,
    IN PWCHAR       Source,
    IN ULONG        Length
    )
{
    ULONG       Index;
    NTSTATUS    Status;

    ASSERT(Destination != NULL);
    ASSERT(Source != NULL);

    for (Index = 0, Status = STATUS_SUCCESS; Index < Length; Index++)
    {
        Destination[Index] = (CHAR) Source[Index];

        //
        // Check that the wide character fits in a normal character
        // but continue with the conversion anyway
        //
        if ( ((WCHAR) Destination[Index]) != Source[Index])
        {
            KdPrint((DRIVER_NAME ": FsdWcharToChar: Invalid character: %C\n",
                Source[Index]));

            Status = STATUS_OBJECT_NAME_INVALID;
        }
    }

    return Status;
}

#endif // !FSD_RO

NTSTATUS
FsdLockUserBuffer (
    IN PIRP             Irp,
    IN ULONG            Length,
    IN LOCK_OPERATION   Operation
    )
{
    NTSTATUS Status;

    ASSERT(Irp != NULL);

    if (Irp->MdlAddress != NULL)
    {
        return STATUS_SUCCESS;
    }

    IoAllocateMdl(Irp->UserBuffer, Length, FALSE, FALSE, Irp);

    if (Irp->MdlAddress == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    __try
    {
        MmProbeAndLockPages(Irp->MdlAddress, Irp->RequestorMode, Operation);

        Status = STATUS_SUCCESS;
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        KdPrint((DRIVER_NAME ": FsdLockUserBuffer: Invalid User Buffer\n"));

        IoFreeMdl(Irp->MdlAddress);

        Irp->MdlAddress = NULL;

        Status = STATUS_INVALID_USER_BUFFER;
    }

    return Status;
}

PVOID
FsdGetUserBuffer (
    IN PIRP Irp
    )
{
    ASSERT(Irp != NULL);

    if (Irp->MdlAddress != NULL)
    {
        return MmGetSystemAddressForMdl(Irp->MdlAddress);
    }
    else
    {
        return Irp->UserBuffer;
    }
}

#pragma code_seg(FSD_PAGED_CODE)

NTSTATUS
FsdReadInodeByIndex (
    IN PDEVICE_OBJECT           DeviceObject,
    IN ULONG                    Index,
    IN OUT struct romfs_inode*  Inode
    )
{
    PUCHAR          Buffer;
    LARGE_INTEGER   Offset;
    NTSTATUS        Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);
    ASSERT(Inode != NULL);

    Buffer = (PUCHAR) FsdAllocatePool(NonPagedPoolCacheAligned, 1024);

    if (Buffer == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Byte offset to the inode
    Index = be32_to_cpu(Index) & ROMFH_MASK;

    KdPrint((DRIVER_NAME ": FsdReadInodeByIndex: Index 0x%x\n", Index));

    // Sector align the offset
    Offset.QuadPart = Index & ~(512-1);

    Status = FsdReadBlockDevice(
        DeviceObject,
        &Offset,
        1024,
        Buffer
        );

    RtlCopyMemory(
        Inode,
        Buffer + (Index & (512-1)),
        sizeof(struct romfs_inode) + ROMFS_MAXFN
        );

    FsdFreePool(Buffer, 1024);

    return Status;
}

NTSTATUS
FsdReadFileData (
    IN PDEVICE_OBJECT       DeviceObject,
    IN ULONG                Index,
    IN struct romfs_inode*  Inode,
    IN PLARGE_INTEGER       Offset,
    IN ULONG                Length,
    IN OUT PVOID            Buffer
    )
{
    LARGE_INTEGER   PhysicalOffset;
    ULONG           PhysicalLength;
    PUCHAR          PhysicalBuffer;
    NTSTATUS        Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);
    ASSERT(Inode != NULL);
    ASSERT(Offset != NULL);
    ASSERT(Buffer != NULL);

    KdPrint((DRIVER_NAME
        ": FsdReadFileData: Index 0x%x, Offset 0x%I64x, Length 0x%x\n",
        Index, Offset->QuadPart, Length));

    PhysicalOffset.QuadPart =
        (Index + SIZEOF_INODE(Inode) + Offset->QuadPart) & ~(512-1);

    PhysicalLength = (Length & ~(512-1)) + 1024;

    PhysicalBuffer = (PUCHAR) FsdAllocatePool(
        NonPagedPoolCacheAligned,
        PhysicalLength
        );

    if (PhysicalBuffer == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    Status = FsdReadBlockDevice(
        DeviceObject,
        &PhysicalOffset,
        PhysicalLength,
        PhysicalBuffer
        );

    RtlCopyMemory(
        Buffer,
        PhysicalBuffer +
        ((Index + SIZEOF_INODE(Inode) + Offset->QuadPart) & (512-1)),
        Length
        );

    FsdFreePool(PhysicalBuffer, PhysicalLength);

    return Status;
}

#pragma code_seg()
